#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import sys
import enum
import os.path
import random
import logging
import collections

import lbuild.module
import lbuild.environment

import lbuild.exception as le

from .node import BaseNode
from .config import ConfigNode
from .facade import BuildLogFacade

from . import repository
from . import utils

LOGGER = logging.getLogger('lbuild.parser')


class Runner:

    def __init__(self, node, env):
        self.node = node
        self.env = env

    def validate(self):
        if hasattr(self.node, "validate"):
            self.env.stage = Parser.Stage.VALIDATE
            self.node.validate(self.env)

    def build(self):
        self.env.stage = Parser.Stage.BUILD
        self.node.build(self.env)

    def post_build(self, buildlog):
        if hasattr(self.node, "post_build"):
            self.env.stage = Parser.Stage.POST_BUILD
            self.node.post_build(self.env)


class Parser(BaseNode):

    @enum.unique
    class Stage(enum.IntEnum):
        """This order of repo/module function calls"""
        INIT = 1
        PREPARE = 2
        VALIDATE = 3
        BUILD = 4
        POST_BUILD = 5

    def __init__(self, config=None):
        BaseNode.__init__(self, "lbuild", BaseNode.Type.PARSER)
        self._config = config if config else ConfigNode()
        self._config_flat = self._config.flatten()

    @property
    def config(self):
        return self._config_flat

    @property
    def modules(self):
        return {m.fullname:m for m in self.all_modules()}

    @property
    def repositories(self):
        return {r.fullname:r for r in self._findall(BaseNode.Type.REPOSITORY)}

    @property
    def repo_options(self):
        return {o.fullname:o for o in self.all_options(depth=3)}

    @property
    def module_options(self):
        return {o.fullname:o for o in self.all_options() if o.depth > 2}

    def load_repositories(self, repofilenames=None):
        repofiles = set(os.path.realpath(p) for p in utils.listify(repofilenames))
        parsed = set()

        while True:
            # flatten configuration and collect repositories
            self._config_flat = self._config.flatten()
            repofiles = (repofiles | set(self._config_flat.repositories)) - parsed
            if not len(repofiles) + len(parsed):
                raise le.LbuildConfigNoReposException(self)
            parsed |= repofiles

            # Parse only new repositories
            for repofile in repofiles:
                repo = self.parse_repository(repofile)

            # nothing more to extend
            if not self._config_flat._extends:
                break

            for filename, aliases in self._config_flat._extends.items():
                node = self._config.find(filename)
                for alias in aliases:
                    fconfig = self.find_any(alias, types=BaseNode.Type.CONFIG)
                    if not fconfig:
                        raise le.LbuildConfigAliasNotFoundException(self, alias)
                    if len(fconfig) > 1:
                        raise le.LbuildConfigAliasAmbiguousException(self, alias, fconfig)
                    self._config.extend(node, ConfigNode.from_file(fconfig[0]._config))
                node._extends.pop(filename)

        self._update_format()
        LOGGER.info("\n%s", self._config.render())
        return self._config_flat

    def parse_repository(self, repofilename: str) -> repository.Repository:
        """
        Parse the given repository file.

        Executes the 'prepare' function to populate the repository
        structure.
        """
        repo = repository.load_repository_from_file(self, repofilename)
        conflict = next((c for c in self.children if c.name == repo.name), None)
        if conflict is not None:
            raise le.LbuildParserDuplicateRepoException(self, repo, conflict)
        repo.parent = self

        return repo

    def merge_repository_options(self):
        # only deal with repo options that contain one `:`
        resolver = self.option_resolver
        for name, value in {n: v for n, v in self.config.options.items()
                            if n.count(":") == 1}.items():
            try:
                resolver[name].value = value
            except le.LbuildOptionException as error:
                raise le.LbuildDumpConfigException(
                    "Failed to validate repository options!\n{}".format(error), self)
            except le.LbuildResolverNoMatchException as error:
                raise le.LbuildParserNodeNotFoundException(self, name, self.Type.OPTION)


    def prepare_repositories(self):
        undefined = self._undefined_repo_options()
        if undefined:
            raise le.LbuildOptionRequiredInputsException(undefined)

        modules = []
        for repo in self._findall(BaseNode.Type.REPOSITORY):
            moduleinits = repo.prepare()
            if not moduleinits:
                raise le.LbuildParserRepositoryEmptyException(repo)
            modules.extend(moduleinits)

        modules = lbuild.module.build_modules(modules)
        # Update the formatters now from repo to the module leaves
        self._update_format()
        return modules

    def merge_module_options(self):
        # only deal with repo options that contain one `:`
        resolver = self.option_resolver
        for name, value in {n: v for n, v in self.config.options.items()
                            if n.count(":") > 1}.items():
            try:
                resolver[name].value = value
            except le.LbuildOptionException as error:
                raise le.LbuildDumpConfigException(
                    "Failed to validate module options!\n{}".format(error), self)
            except le.LbuildResolverNoMatchException as error:
                raise le.LbuildParserNodeNotFoundException(self, name, self.Type.OPTION)

    def resolve_dependencies(self, requested_modules, depth=sys.maxsize):
        """
        Resolve dependencies by adding missing modules.

        Args:
            modules (list): All modules available in the system
            requested_modules (list): Selected modules. Sub-set of all modules.
            depth (int): Maximum depth of dependencies. If not set to maximum
                not all dependencies might be resolved.

        Returns:
            list: Required modules for the given list of modules.
        """
        selected_modules = set(requested_modules.copy())

        LOGGER.info("Selected modules: %s",
                    ", ".join(sorted(module.fullname for module in selected_modules)))

        try:
            current = selected_modules
            while depth > 0:
                additional = set()
                for module in current:
                    for dependency in module.dependencies:
                        if dependency not in selected_modules and dependency not in additional:
                            LOGGER.debug("Adding dependency: %s", dependency.fullname)
                            additional.add(dependency)
                if not additional:
                    # Abort if no new dependencies are being found
                    break
                selected_modules.update(additional)
                current = additional
                additional = []
                depth -= 1

        except le.LbuildResolverNoMatchException as error:
            raise le.LbuildParserCannotResolveDependencyException(self, error)

        # disable all non-selected modules
        for module in self.all_modules():
            if module._available:
                module._selected = module in selected_modules

        self._update()
        return selected_modules

    def find_module(self, name):
        return self.module_resolver[name]

    def find_option(self, name):
        return self.option_resolver[name]

    def find_modules(self, queries):
        return self.find_any(queries, self.Type.MODULE)

    def find_all(self, queries):
        return [node for q in utils.listify(queries) for node in self.find_any(q)]

    def find_any(self, queries, types=None):
        nodes = set()
        for query in utils.listify(queries):
            result = self._resolve_partial(query, None)
            if result is None:
                raise le.LbuildParserNodeNotFoundException(self, query, types)
            nodes |= set(result)
        if types:
            types = utils.listify(types)
            nodes = [n for n in nodes if any(n.type == t for t in types)]
        return list(nodes)

    @staticmethod
    def _undefined_options(options):
        return {o for o in options if o.value is None}

    def _undefined_repo_options(self):
        return self._undefined_options(self.repo_options.values())

    def validate_modules(self, build_modules):
        self.build_modules(build_modules, None)

    def build_modules(self, build_modules, buildlog):
        undefined = self._undefined_options(self.all_options())
        if undefined:
            raise le.LbuildOptionRequiredInputsException(undefined)
        if not build_modules:
            raise le.LbuildConfigNoModulesException(self)

        groups = collections.defaultdict(list)
        # Build environments for all modules and repos
        for node in (build_modules | {m.repository for m in build_modules}):
            env = lbuild.environment.Environment(node, buildlog)
            groups[node.depth].append(Runner(node, env))

        exceptions = []
        # Enforce that the submodules are always build before their
        # parent modules.
        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                try:
                    runner.validate()
                except le.LbuildException as error:
                    exceptions.append(error)

        if exceptions:
            raise le.LbuildAggregateException(exceptions)

        # Cannot build with these settings, just pre_build
        if buildlog is None:
            return

        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                runner.build()

        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                runner.post_build(buildlog)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
from os.path import realpath, join
import pkgutil
import logging
import collections
from pathlib import Path

import lxml.etree
import anytree

from .exception import LbuildConfigException, LbuildConfigNotFoundException

LOGGER = logging.getLogger('lbuild.config')
DEFAULT_CACHE_FOLDER = ".lbuild_cache"


class ConfigNode(anytree.AnyNode):

    def __init__(self, parent=None):
        anytree.AnyNode.__init__(self, parent)

        self._cachefolder = Path()
        self._extends = collections.defaultdict(list)
        self._vcs = []
        self._repositories = []
        self._modules = []
        self._options = {}

        self.filename = Path()

    @property
    def repositories(self):
        return self._repositories

    @property
    def modules(self):
        return self._modules

    @property
    def options(self):
        return self._options

    @property
    def vcs(self):
        return self._vcs

    @property
    def cachefolder(self):
        return self._cachefolder

    def add_commandline_options(self, cmd_options):
        for option in cmd_options:
            parts = option.split('=')
            self._options[parts[0]] = parts[1]

    def _flatten(self, config):
        for node in list(self.siblings) + [self]:
            config._cachefolder = node._cachefolder
            config._extends.update(node._extends)
            config._vcs.extend(node._vcs)
            config._repositories.extend(node._repositories)
            config._modules.extend(node._modules)
            config._options.update(node._options)

        if self.parent:
            self.parent._flatten(config)

    def flatten(self):
        config = ConfigNode()
        self.last._flatten(config)
        config._repositories = list(set(config._repositories))
        config._modules = list(set(config._modules))
        return config

    @staticmethod
    def extend(node, config):
        if config and node:
            below = next(iter(node.children), None)
            config.parent = node
            if below is not None:
                below.parent = config
        return node

    def extend_last(self, config):
        return self.extend(self.last, config)

    @property
    def last(self):
        descendants = self.root.descendants
        return descendants[-1] if descendants else self

    def find(self, filename):
        return anytree.find_by_attr(self.root, name="filename", value=filename)

    def render(self):
        if self.filename == Path():
            return "ConfigNode(command-line)"
        return anytree.RenderTree(self, anytree.ContRoundStyle())

    @staticmethod
    def from_path(startpath=None, name="lbuild.xml"):
        """
        Iterate upwards from the starting folder to find all
        configuration files.

        Args:
            startpath -- Start point for the iteration. If omitted
                the current working directory is used.
            name -- Filename of the configuation file.

        Returns:
            `ConfigNode` if a configuration file was found, `None`
            otherwise.
        """
        configs = []
        startpath = Path(startpath) if startpath else Path.cwd()
        while startpath.exists():
            config = (startpath / name)
            if config.exists():
                configs.append(ConfigNode.from_file(config))
            if startpath.parent == startpath:
                break
            startpath = startpath.parent

        if not configs:
            return None
        if len(configs) == 1:
            return configs[0]

        first = configs[0]
        for node in configs[1:]:
            first.extend_last(node)
        return first

    @staticmethod
    def from_file(configfile, parent=None):
        """
        Load configuration from an XML configuration file.

        Args:
            configfile -- Path to the configuration file.
            parent -- Exisiting configuration which will be updated.

        Returns:
            Configuration as a `ConfigNode` instance.
        """
        filename = realpath(str(configfile))
        if not os.path.exists(filename):
            raise LbuildConfigNotFoundException(filename)

        xmltree = ConfigNode._load_and_verify(configfile)

        config = ConfigNode(parent)
        config.filename = os.path.relpath(filename, os.getcwd())
        LOGGER.debug("Parse configuration '%s'", config.filename)
        configpath = os.path.dirname(config.filename)
        # load extend strings
        for node in xmltree.iterfind("extends"):
            cpath = Path(ConfigNode._rel_path(node.text, configpath))
            # We first need to check for the path, because in Windows `:`
            # is allowed to be part of the path (`C:\path`)
            try:
                cpath_exists = cpath.exists()
            except:
                cpath_exists = False
            if cpath_exists:
                ConfigNode.from_file(str(cpath), config)
            elif ":" in node.text:
                config._extends[config.filename].append(node.text)
            else:
                raise LbuildConfigNotFoundException(cpath, filename)


        # Load cachefolder
        cache_node = xmltree.find("repositories/cache")
        if cache_node is not None:
            config._cachefolder = ConfigNode._rel_path(cache_node.text, configpath)
        else:
            config._cachefolder = join(configpath, DEFAULT_CACHE_FOLDER)

        # Load version control nodes
        for vcs_node in xmltree.iterfind("repositories/repository/vcs"):
            config._vcs += [ConfigNode.to_dict(vcs) for vcs in vcs_node.iterchildren()]

        # Load repositories
        for path_node in xmltree.iterfind("repositories/repository/path"):
            repopath = path_node.text.format(cache=config._cachefolder)
            filename = ConfigNode._rel_path(repopath, configpath)
            config._repositories.append(filename)

        # Load all requested modules
        config._modules = xmltree.xpath('modules/module/text()')

        # Load options
        for option_node in xmltree.xpath('options/option'):
            name = option_node.attrib['name']
            value = option_node.attrib.get('value', option_node.text)
            config._options[name] = value

        return config

    @staticmethod
    def _load_and_verify(configfile):
        try:
            xmlroot = lxml.etree.parse(str(configfile))

            xmlschema = lxml.etree.fromstring(
                pkgutil.get_data('lbuild', 'resources/configuration.xsd'))
            schema = lxml.etree.XMLSchema(xmlschema)
            schema.assertValid(xmlroot)

            xmltree = xmlroot.getroot()
        except OSError as error:
            raise LbuildConfigException(configfile, error)
        except (lxml.etree.DocumentInvalid,
                lxml.etree.XMLSyntaxError,
                lxml.etree.XMLSchemaParseError,
                lxml.etree.XIncludeError) as error:
            # lxml.etree has the used exception, but pylint is not able to detect them:
            # pylint: disable=no-member
            raise LbuildConfigException(configfile, error)
        return xmltree

    @staticmethod
    def _rel_path(path, configpath):
        if not os.path.isabs(path):
            path = join(configpath, path)
        return realpath(path)

    @staticmethod
    def to_dict(xmltree):
        """
        Convert XML to a Python dictionary according to
        http://www.xml.com/pub/a/2006/05/31/converting-between-xml-and-json.html
        """
        root_dict = {xmltree.tag: {} if xmltree.attrib else None}

        children = []
        for node in xmltree:
            children.append(node)

        if children:
            dd = collections.defaultdict(list)
            for dc in [ConfigNode.to_dict(node) for node in children]:
                for key, value in dc.items():
                    dd[key].append(value)
            root_dict = {xmltree.tag: {key: value[0] if len(value) == 1 else value
                                       for key, value in dd.items()}}

        if xmltree.attrib:
            root_dict[xmltree.tag].update(
                ('@' + key, value) for key, value in xmltree.attrib.items())

        if xmltree.text:
            text = xmltree.text.strip()
            if children or xmltree.attrib:
                if text:
                    root_dict[xmltree.tag]['#text'] = text
            else:
                root_dict[xmltree.tag] = text

        return root_dict



#!/usr/bin/env python3
#
# Copyright (c) 2015-2016, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import textwrap
import lbuild.utils


def wordwrap(value, width=79):
    return '\n\n'.join([textwrap.fill(s, width) for s in value.split('\n\n')])


def indent(text, spaces=0, first_line=False):
    """
    Indent the text by the given number of white spaces.
    """
    prefix = ' ' * spaces
    if first_line:
        return textwrap.indent(text, prefix)

    return ("\n" + prefix).join(text.splitlines())


def pad(text, min_width, tabwidth=4):
    """
    Fill the text with tabs at the end until the minimal width is reached.
    """
    tab_count = ((min_width / tabwidth) - (len(text) / tabwidth)) + 1
    return text + ('\t' * int(tab_count))


def split(value, delimiter):
    return value.split(delimiter)


def values(dictionaries, key):
    """
    Go through the list of dictionaries and add all values of a certain key
    to a list.
    """
    value_list = []
    for item in dictionaries:
        if isinstance(item, dict) and key in item:
            if item[key] not in value_list:
                value_list.append(item[key])
    return value_list


def listify(*node):
    return lbuild.utils.listify(*node)


DEFAULT_FILTERS = {
    'lbuild.wordwrap': wordwrap,
    'lbuild.indent': indent,
    'lbuild.pad': pad,
    'lbuild.values': values,
    'lbuild.split': split,
    'lbuild.listify': listify,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, 2017-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import sys
import uuid
import shutil
import fnmatch
import importlib.util
import importlib.machinery

import lbuild.exception as le

DEFAULT_IGNORE_PATTERNS = [
    "*/.git*",
    "*/.DS_Store*",
    "*/__pycache__",
    "*/module.lb",
    "*/repo.lb",
    "*/*.pyc"
]


def ignore_files(*files):
    """
    Ignore file and folder names without checking the full path.

    Example: the following code with ignore all files with the ending `.lb`:
    ```
    env.copy(".", ignore=env.ignore_files("*.lb"))
    ```

    Based on the shutil.ignore_patterns() function.
    """
    return shutil.ignore_patterns(*files)


def ignore_patterns(*patterns):
    """
    Ignore patterns based on the absolute file path.

    Use an `*` at the beginning to match relative paths:
    ```
    env.copy(".", ignore=env.ignore_patterns("*platform/*.lb"))
    ```
    This ignores all files in the `platform` sub-directory with the
    ending `.lb`.
    """

    def check(path, files):
        ignored = set()
        for pattern in patterns:
            for filename in files:
                if fnmatch.fnmatch(os.path.join(path, filename), pattern):
                    # The copytree function uses only the filename to check
                    # which files should be ignored, not the absolute path.
                    ignored.add(filename)
        return ignored

    return check


def _listify(obj):
    if obj is None:
        return list()
    if isinstance(obj, (list, tuple, set, range)):
        return list(obj)
    if hasattr(obj, "__iter__") and not hasattr(obj, "__getitem__"):
        return list(obj)
    return [obj, ]


def listify(*objs):
    """
    Convert arguments to list if they are not already a list.
    """
    return [l for o in objs for l in _listify(o)]


def listrify(*objs):
    """
    Convert arguments to list of strings.
    """
    return list(map(str, listify(*objs)))


def get_global_functions(env, required, optional=None):
    """
    Get global functions an environment.

    Args:
        required: List of required functions.
        optional: List of optional functions.
    """

    def get(name, fail=True):
        if isinstance(env, dict):
            val = env.get(name, None)
        else:
            val = getattr(env, name, None)
        if fail and val is None:
            raise le.LbuildUtilsFunctionNotFoundException(name, required, optional)
        return val

    if optional is None:
        optional = []

    functions = {}
    for name in required + optional:
        function = get(name, name in required)
        if function is not None:
            functions[name] = function

    return functions


def load_module_from_file(filename, local, modulename=None):
    """
    Load a python module from a local file.

    Args:
        filename: Name of the file to load. The file extension is ignored.
        local: dictionary of symbols which will be added to the global
            namespace when executing the module code.
        modulename: Name of the module. When set to `None`.

    Returns:
        Namespace of the module.
    """
    # The actual name of the module is only known after it is loaded. Therefore
    # a UUID is used instead here.
    if modulename is None:
        modulename = "lbuild.modules.{}".format(uuid.uuid1())

    loader = importlib.machinery.SourceFileLoader(modulename, filename)

    spec = importlib.util.spec_from_loader(loader.name, loader)
    try:
        module = importlib.util.module_from_spec(spec)
    except AttributeError:
        import types
        module = types.ModuleType(modulename)

    # Prepare the environment of the module. Everything set here will
    # be available in the global namespace when executing the module.
    module.__dict__.update(local)

    try:
        # Load the module. This executes the code inside the lbuild module file.
        loader.exec_module(module)
    except Exception as error:
        raise le.LbuildForwardException(modulename, error)

    sys.modules[modulename] = module
    return module.__dict__


def with_forward_exception(module, function):
    """
    Run a function a store exceptions as forward exceptions.
    """
    try:
        return function()
    except le.LbuildException:
        raise
    except Exception as error:
        raise le.LbuildForwardException(module, error)


import errno
import tempfile

def is_pathname_valid(path):
    if not isinstance(path, str) or not path:
        return False
    if (os.path.sep + os.path.sep) in path:
        return False
    return _is_pathname_valid(path)

# See https://stackoverflow.com/a/34102855
def _is_pathname_valid(pathname: str) -> bool:
    try:
        _, pathname = os.path.splitdrive(pathname)
        with tempfile.TemporaryDirectory() as root_dirname:
            root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep
            for pathname_part in pathname.split(os.path.sep):
                try:
                    os.lstat(root_dirname + pathname_part)
                except OSError as exc:
                    if hasattr(exc, "winerror"):
                        if exc.winerror == 123:
                            return False
                    elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                        return False
    except TypeError as exc:
        return False
    return True
